use crate::base::{ErrorCode, HasErrorCode};
use crate::{
    error, ErrorCodeReaderExt, ErrorCodeWriterExt, HasFixedCommandId,
    IsRequest, IsResponse, IsValidResponseFor, MessageExt,
    ReadFromPayload, Result, TryFromResponse, WithFixedPayloadLength,
    WriteExt,
};
use snafu::ResultExt;
use std::io::{Read, Write};

#[derive(Clone, Debug, PartialEq)]
pub struct Request;

impl IsRequest for Request {
    type Response = Response;
}

impl From<Request> for super::Request {
    fn from(m: Request) -> Self {
        super::Request::QueryDeviceManufacturer(m)
    }
}

impl WithFixedPayloadLength for Request {
    const FIXED_PAYLOAD_LENGTH: u16 = 0u16;
}

impl HasFixedCommandId for Request {
    const COMMAND_ID: u16 = 0x0802;
}

impl MessageExt for Request {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, _w: &mut dyn Write) -> Result<()> {
        Ok(())
    }
}

impl ReadFromPayload for Request {
    fn read_from_payload<R: Read>(
        _r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_device_firmware_version_response",
        )?;
        Ok(Self {})
    }
}

#[cfg(test)]
mod test_request {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Request {};
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x02, // command lower byte
                0x08, // command upper byte
                0x00, // length lower byte
                0x00, // length upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![];
        let len = buffer.len() as u16;
        let message =
            Request::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(message, Request {});
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Response {
    pub error: ErrorCode,
    pub manufacturer: String,
}

impl IsResponse for Response {
    type Request = Request;
}

impl IsValidResponseFor for Response {
    fn is_valid_response_for(&self, _request: &Request) -> bool {
        true
    }
}

impl HasErrorCode for Response {
    fn error_code(&self) -> ErrorCode {
        self.error
    }
}

impl TryFromResponse for Response {
    fn try_from_response(r: super::Response) -> Result<Self> {
        if let super::Response::QueryDeviceManufacturer(r) = r {
            Ok(r)
        } else {
            error::UnexpectedResponseType.fail()?
        }
    }
}

impl From<Response> for super::Response {
    fn from(m: Response) -> Self {
        super::Response::QueryDeviceManufacturer(m)
    }
}

impl HasFixedCommandId for Response {
    const COMMAND_ID: u16 = 0x0803;
}

impl MessageExt for Response {
    fn payload_length(&self) -> u16 {
        ErrorCode::FIXED_PAYLOAD_LENGTH
            + self.manufacturer.as_bytes().len() as u16
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_error_code(self.error)?;
        w.write_all_(self.manufacturer.as_bytes())?;
        Ok(())
    }
}

impl ReadFromPayload for Response {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        let error = r.read_error_code()?;
        let mut manufacturer = vec![
            0u8;
            (payload_length - ErrorCode::FIXED_PAYLOAD_LENGTH)
                as usize
        ];
        r.read_exact(&mut manufacturer).context(error::Io)?;
        let manufacturer = String::from_utf8(manufacturer)
            .context(error::Utf8Conversion)?;
        Ok(Self {
            error,
            manufacturer,
        })
    }
}

#[cfg(test)]
mod test_response {
    use super::*;

    #[test]
    fn write_to_empty() {
        let mut buffer = vec![];
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let error = ErrorCode::Success;
        let manufacturer = "".to_string();

        let message = Response {
            error,
            manufacturer,
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            vec![
                0x45, // options
                0x93, // sequence
                0x03, // command lower byte
                0x08, // command upper byte
                0x02, // length lower byte
                0x00, // length upper byte
                0x00, // error lower byte
                0x00, // error upper byte
            ]
        );
    }
    #[test]
    fn write_to_name() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let error = ErrorCode::Success;
        let manufacturer = "ACompany".to_string();

        let message = Response {
            error,
            manufacturer,
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            vec![
                0x45, // options
                0x93, // sequence
                0x03, // command lower byte
                0x08, // command upper byte
                0x0A, // length lower byte
                0x00, // length upper byte
                0x00, // error lower byte
                0x00, // error upper byte
                0x41, // 'A'
                0x43, // 'C'
                0x6F, // 'o'
                0x6D, // 'm'
                0x70, // 'p'
                0x61, // 'a'
                0x6E, // 'n'
                0x79, // 'y'
            ]
        );
    }
    #[test]
    fn from_payload_empty() {
        let buffer = vec![
            0x00, // error lower byte
            0x00, // error upper byte
        ];
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();

        let error = ErrorCode::Success;
        let manufacturer = "".to_string();

        assert_eq!(
            message,
            Response {
                error,
                manufacturer,
            }
        );
    }
    #[test]
    fn from_payload_name() {
        let buffer = vec![
            0x00, // error lower byte
            0x00, // error upper byte
            0x41, // 'A'
            0x43, // 'C'
            0x6F, // 'o'
            0x6D, // 'm'
            0x70, // 'p'
            0x61, // 'a'
            0x6E, // 'n'
            0x79, // 'y'
        ];
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();

        let error = ErrorCode::Success;
        let manufacturer = "ACompany".to_string();

        assert_eq!(
            message,
            Response {
                error,
                manufacturer,
            }
        );
    }
}
