use crate::base::{ErrorCode, HasErrorCode};
use crate::{
    error, ErrorCodeReaderExt, ErrorCodeWriterExt, HasFixedCommandId,
    IsRequest, IsResponse, IsValidResponseFor, MessageExt,
    ReadFromPayload, Result, TryFromResponse, WithFixedPayloadLength,
    WriteExt,
};
use snafu::ResultExt;
use std::io::{Read, Write};
use uuid::{Bytes, Uuid};

#[derive(Clone, Debug, PartialEq)]
pub struct Request;

impl IsRequest for Request {
    type Response = Response;
}

impl From<Request> for super::Request {
    fn from(m: Request) -> Self {
        super::Request::QueryDeviceUuid(m)
    }
}

impl WithFixedPayloadLength for Request {
    const FIXED_PAYLOAD_LENGTH: u16 = 0u16;
}

impl HasFixedCommandId for Request {
    const COMMAND_ID: u16 = 0x0804;
}

impl MessageExt for Request {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, _w: &mut dyn Write) -> Result<()> {
        Ok(())
    }
}

impl ReadFromPayload for Request {
    fn read_from_payload<R: Read>(
        _r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_device_uuid_request",
        )?;
        Ok(Self {})
    }
}

#[cfg(test)]
mod test_request {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Request {};
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x04, // command lower byte
                0x08, // command upper byte
                0x00, // length lower byte
                0x00, // length upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![];
        let len = buffer.len() as u16;
        let message =
            Request::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(message, Request {});
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Response {
    pub error: ErrorCode,
    pub uuid: Uuid,
}

impl IsResponse for Response {
    type Request = Request;
}

impl IsValidResponseFor for Response {
    fn is_valid_response_for(&self, _request: &Request) -> bool {
        true
    }
}

impl HasErrorCode for Response {
    fn error_code(&self) -> ErrorCode {
        self.error
    }
}

impl TryFromResponse for Response {
    fn try_from_response(r: super::Response) -> Result<Self> {
        if let super::Response::QueryDeviceUuid(r) = r {
            Ok(r)
        } else {
            error::UnexpectedResponseType.fail()?
        }
    }
}

impl From<Response> for super::Response {
    fn from(m: Response) -> Self {
        super::Response::QueryDeviceUuid(m)
    }
}

impl WithFixedPayloadLength for Response {
    const FIXED_PAYLOAD_LENGTH: u16 =
        ErrorCode::FIXED_PAYLOAD_LENGTH + Uuid::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Response {
    const COMMAND_ID: u16 = 0x0805;
}

impl MessageExt for Response {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_error_code(self.error)?;
        w.write_all_(self.uuid.as_bytes())?;
        Ok(())
    }
}

impl ReadFromPayload for Response {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_device_uuid_response",
        )?;
        let error = r.read_error_code()?;

        let mut bytes: Bytes = [0u8; 16];
        r.read_exact(&mut bytes).context(error::Io)?;
        let uuid = Uuid::from_bytes(bytes);
        Ok(Self { error, uuid })
    }
}

#[cfg(test)]
mod test_response {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let uuid = Uuid::parse_str("ad84895b-7872-42f6-97ff-e4920cbe86d5")
            .unwrap();

        let message = Response {
            error: ErrorCode::InvalidState, // 0x0107
            uuid,
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x05, // command lower byte
                0x08, // command upper byte
                0x12, // length lower byte
                0x00, // length upper byte
                0x07, // error lower byte
                0x01, // error upper byte
                0xAD, // uuid
                0x84, // uuid
                0x89, // uuid
                0x5b, // uuid
                0x78, // uuid
                0x72, // uuid
                0x42, // uuid
                0xF6, // uuid
                0x97, // uuid
                0xFF, // uuid
                0xE4, // uuid
                0x92, // uuid
                0x0C, // uuid
                0xBE, // uuid
                0x86, // uuid
                0xD5, // uuid
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![
            0x07, // error lower byte
            0x01, // error upper byte
            0xAD, // uuid
            0x84, // uuid
            0x89, // uuid
            0x5b, // uuid
            0x78, // uuid
            0x72, // uuid
            0x42, // uuid
            0xF6, // uuid
            0x97, // uuid
            0xFF, // uuid
            0xE4, // uuid
            0x92, // uuid
            0x0C, // uuid
            0xBE, // uuid
            0x86, // uuid
            0xD5, // uuid
        ];
        let uuid = Uuid::parse_str("ad84895b-7872-42f6-97ff-e4920cbe86d5")
            .unwrap();
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(
            message,
            Response {
                error: ErrorCode::InvalidState,
                uuid,
            }
        );
    }
}
