use crate::base::{ErrorCode, HasErrorCode};
use crate::{
    error, ErrorCodeReaderExt, ErrorCodeWriterExt, HasFixedCommandId,
    IsRequest, IsResponse, IsValidResponseFor, MessageExt, ReadExt,
    ReadFromPayload, Result, TryFromResponse, WithFixedPayloadLength,
    WriteExt,
};
use std::io::{Read, Write};

#[derive(Clone, Debug, PartialEq)]
pub struct Request;

impl IsRequest for Request {
    type Response = Response;
}

impl From<Request> for super::Request {
    fn from(m: Request) -> Self {
        super::Request::QueryFrameFormatVersion(m)
    }
}

impl WithFixedPayloadLength for Request {
    const FIXED_PAYLOAD_LENGTH: u16 = 0u16;
}

impl HasFixedCommandId for Request {
    const COMMAND_ID: u16 = 0x0000;
}

impl MessageExt for Request {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, _w: &mut dyn Write) -> Result<()> {
        Ok(())
    }
}

impl ReadFromPayload for Request {
    fn read_from_payload<R: Read>(
        _r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_frame_format_version_request",
        )?;
        Ok(Self {})
    }
}

#[cfg(test)]
mod test_request {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Request {};
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x00, // command lower byte
                0x00, // command upper byte
                0x00, // length lower byte
                0x00, // length upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![];
        let len = buffer.len() as u16;
        let message =
            Request::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(message, Request {});
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Response {
    pub error: ErrorCode,
    pub version: u8,
}

impl IsResponse for Response {
    type Request = Request;
}

impl IsValidResponseFor for Response {
    fn is_valid_response_for(&self, _request: &Request) -> bool {
        true
    }
}

impl HasErrorCode for Response {
    fn error_code(&self) -> ErrorCode {
        self.error
    }
}

impl TryFromResponse for Response {
    fn try_from_response(r: super::Response) -> Result<Self> {
        if let super::Response::QueryFrameFormatVersion(r) = r {
            Ok(r)
        } else {
            error::UnexpectedResponseType.fail()?
        }
    }
}

impl From<Response> for super::Response {
    fn from(m: Response) -> Self {
        super::Response::QueryFrameFormatVersion(m)
    }
}

impl WithFixedPayloadLength for Response {
    const FIXED_PAYLOAD_LENGTH: u16 =
        ErrorCode::FIXED_PAYLOAD_LENGTH + u8::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Response {
    const COMMAND_ID: u16 = 0x0001;
}

impl MessageExt for Response {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_error_code(self.error)?;
        w.write_u8(self.version)?;
        Ok(())
    }
}

impl ReadFromPayload for Response {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_frame_format_version_response",
        )?;
        let error = r.read_error_code()?;
        let version = r.read_u8()?;
        Ok(Self { error, version })
    }
}

#[cfg(test)]
mod test_response {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Response {
            error: ErrorCode::InvalidState, // 0x0107
            version: 0x23,
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x01, // command lower byte
                0x00, // command upper byte
                0x03, // length lower byte
                0x00, // length upper byte
                0x07, // error lower byte
                0x01, // error upper byte
                0x23, // version
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![
            0x07, // error lower byte
            0x01, // error upper byte
            0x23, // version
        ];
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(
            message,
            Response {
                error: ErrorCode::InvalidState,
                version: 0x23,
            }
        );
    }
}
