use crate::base::{ErrorCode, HasErrorCode, ModuleState};
use crate::{
    error, Endian, ErrorCodeReaderExt, ErrorCodeWriterExt,
    HasFixedCommandId, IsRequest, IsResponse, IsValidResponseFor,
    MessageExt, ModuleStateReaderExt, ModuleStateWriterExt, ReadExt,
    ReadFromPayload, Result, TryFromResponse, WithFixedPayloadLength,
    WriteExt,
};
use std::io::{Read, Write};

#[derive(Clone, Debug, PartialEq)]
pub struct Request {
    pub module_id: u16,
}

impl IsRequest for Request {
    type Response = Response;
}

impl From<Request> for super::Request {
    fn from(m: Request) -> Self {
        super::Request::QueryModuleState(m)
    }
}

impl WithFixedPayloadLength for Request {
    const FIXED_PAYLOAD_LENGTH: u16 = u16::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Request {
    const COMMAND_ID: u16 = 0x380c;
}

impl MessageExt for Request {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_u16::<Endian>(self.module_id)?;
        Ok(())
    }
}

impl ReadFromPayload for Request {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_module_state_request",
        )?;
        let module_id = r.read_u16::<Endian>()?;
        Ok(Self { module_id })
    }
}

#[cfg(test)]
mod test_request {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Request { module_id: 0x1234 };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x0c, // command lower byte
                0x38, // command upper byte
                0x02, // length lower byte
                0x00, // length upper byte
                0x34, // module_id lower byte
                0x12, // module_id upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![0x34, 0x12];
        let len = buffer.len() as u16;
        let message =
            Request::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(message, Request { module_id: 0x1234 });
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Response {
    pub error: ErrorCode,
    pub state: ModuleState,
    pub assigned: bool,
    pub module_error: ErrorCode,
}

impl IsResponse for Response {
    type Request = Request;
}

impl IsValidResponseFor for Response {
    fn is_valid_response_for(&self, _request: &Request) -> bool {
        true
    }
}

impl HasErrorCode for Response {
    fn error_code(&self) -> ErrorCode {
        self.error
    }
}

impl TryFromResponse for Response {
    fn try_from_response(r: super::Response) -> Result<Self> {
        if let super::Response::QueryModuleState(r) = r {
            Ok(r)
        } else {
            error::UnexpectedResponseType.fail()?
        }
    }
}

impl From<Response> for super::Response {
    fn from(m: Response) -> Self {
        super::Response::QueryModuleState(m)
    }
}

impl WithFixedPayloadLength for Response {
    const FIXED_PAYLOAD_LENGTH: u16 = ErrorCode::FIXED_PAYLOAD_LENGTH
        + ModuleState::FIXED_PAYLOAD_LENGTH
        + bool::FIXED_PAYLOAD_LENGTH
        + ErrorCode::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Response {
    const COMMAND_ID: u16 = 0x380D;
}

impl MessageExt for Response {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_error_code(self.error)?;
        w.write_module_state(&self.state)?;
        w.write_u8(self.assigned as u8)?;
        w.write_error_code(self.module_error)?;
        Ok(())
    }
}

impl ReadFromPayload for Response {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "query_module_state_response",
        )?;
        let error = r.read_error_code()?;
        let state = r.read_module_state()?;
        let assigned = r.read_u8()? > 0;
        let module_error = r.read_error_code()?;
        Ok(Self {
            error,
            state,
            assigned,
            module_error,
        })
    }
}

#[cfg(test)]
mod test_response {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Response {
            error: ErrorCode::InvalidState, // 0x0107
            state: ModuleState::Running,    // 0x02
            assigned: true,
            module_error: ErrorCode::UnsupportedType, // 0x0102
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x0D, // command lower byte
                0x38, // command upper byte
                0x06, // length lower byte
                0x00, // length upper byte
                0x07, // error lower byte
                0x01, // error upper byte
                0x02, // state
                0x01, // assigned
                0x02, // module_error upper byte
                0x01, // module_error upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![
            0x07, // error lower byte
            0x01, // error upper byte
            0x02, // state
            0x01, // assigned
            0x02, // module_error upper byte
            0x01, // module_error lower byte
        ];
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(
            message,
            Response {
                error: ErrorCode::InvalidState, // 0x0107
                state: ModuleState::Running,    // 0x02
                assigned: true,
                module_error: ErrorCode::UnsupportedType, // 0x0102
            }
        );
    }
}
