use crate::base::{DataValueRaw, ErrorCode, HasErrorCode};
use crate::{
    error, DataValueReaderExt, DataValueWriterExt, Endian,
    ErrorCodeReaderExt, ErrorCodeWriterExt, HasFixedCommandId, IsRequest,
    IsResponse, IsValidResponseFor, MessageExt, ReadExt, ReadFromPayload,
    Result, TryFromResponse, WithFixedPayloadLength, WriteExt,
};
use std::io::{Read, Write};

#[derive(Clone, Debug, PartialEq)]
pub struct Request {
    pub module_id: u16,
    pub channel_index: u16,
}

impl IsRequest for Request {
    type Response = Response;
}

impl From<Request> for super::Request {
    fn from(m: Request) -> Self {
        super::Request::ReadValueFromModuleChannel(m)
    }
}

impl WithFixedPayloadLength for Request {
    const FIXED_PAYLOAD_LENGTH: u16 =
        u16::FIXED_PAYLOAD_LENGTH + u16::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Request {
    const COMMAND_ID: u16 = 0x380a;
}

impl MessageExt for Request {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_u16::<Endian>(self.module_id)?;
        w.write_u16::<Endian>(self.channel_index)?;
        Ok(())
    }
}

impl ReadFromPayload for Request {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "read_value_from_module_channel_request",
        )?;
        let module_id = r.read_u16::<Endian>()?;
        let channel_index = r.read_u16::<Endian>()?;
        Ok(Self {
            module_id,
            channel_index,
        })
    }
}

#[cfg(test)]
mod test_request {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Request {
            module_id: 0x1122,
            channel_index: 0x3344,
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x0a, // command lower byte
                0x38, // command upper byte
                0x04, // length lower byte
                0x00, // length upper byte
                0x22, // module_id lower byte
                0x11, // module_id upper byte
                0x44, // channel_index lower byte
                0x33, // channel_index upper byte
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![
            0x22, // module_id lower byte
            0x11, // module_id upper byte
            0x44, // channel_index lower byte
            0x33, // channel_index upper byte
        ];
        let len = buffer.len() as u16;
        let message =
            Request::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(
            message,
            Request {
                module_id: 0x1122,
                channel_index: 0x3344,
            }
        );
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct Response {
    pub error: ErrorCode,
    pub value: DataValueRaw,
}

impl IsResponse for Response {
    type Request = Request;
}

impl IsValidResponseFor for Response {
    fn is_valid_response_for(&self, _request: &Request) -> bool {
        true
    }
}

impl HasErrorCode for Response {
    fn error_code(&self) -> ErrorCode {
        self.error
    }
}

impl TryFromResponse for Response {
    fn try_from_response(r: super::Response) -> Result<Self> {
        if let super::Response::ReadValueFromModuleChannel(r) = r {
            Ok(r)
        } else {
            error::UnexpectedResponseType.fail()
        }
    }
}

impl From<Response> for super::Response {
    fn from(m: Response) -> Self {
        super::Response::ReadValueFromModuleChannel(m)
    }
}

impl WithFixedPayloadLength for Response {
    const FIXED_PAYLOAD_LENGTH: u16 = ErrorCode::FIXED_PAYLOAD_LENGTH
        + DataValueRaw::FIXED_PAYLOAD_LENGTH;
}

impl HasFixedCommandId for Response {
    const COMMAND_ID: u16 = 0x380b;
}

impl MessageExt for Response {
    fn payload_length(&self) -> u16 {
        Self::FIXED_PAYLOAD_LENGTH
    }

    fn write_payload(&self, w: &mut dyn Write) -> Result<()> {
        w.write_error_code(self.error)?;
        w.write_data_value(&self.value)?;
        Ok(())
    }
}

impl ReadFromPayload for Response {
    fn read_from_payload<R: Read>(
        r: &mut R,
        payload_length: u16,
    ) -> Result<Self> {
        Self::verify_payload_length(
            payload_length,
            "read_value_from_module_channel_repsonse",
        )?;
        let error = r.read_error_code()?;
        let value = r.read_data_value()?;
        Ok(Self { error, value })
    }
}

#[cfg(test)]
mod test_response {
    use super::*;

    #[test]
    fn write_to() {
        let mut buffer = Vec::new();
        let options: u8 = 0x45;
        let sequence: u8 = 0x93;

        let message = Response {
            error: ErrorCode::InvalidState, // 0x0107
            value: 0x6543u16.into(),
        };
        message.write_to(&mut buffer, options, sequence).unwrap();
        assert_eq!(
            buffer,
            [
                0x45, // options
                0x93, // sequence
                0x0b, // command lower byte
                0x38, // command upper byte
                0x0B, // length lower byte
                0x00, // length upper byte
                0x07, // error lower byte
                0x01, // error upper byte
                0x06, // value type
                0x43, // value byte 0
                0x65, // value byte 1
                0x00, // value byte 2
                0x00, // value byte 3
                0x00, // value byte 4
                0x00, // value byte 5
                0x00, // value byte 6
                0x00, // value byte 7
            ]
        );
    }
    #[test]
    fn read_from_payload() {
        let buffer = vec![
            0x07, // error lower byte
            0x01, // error upper byte
            0x06, // value type
            0x43, // value byte 0
            0x65, // value byte 1
            0x00, // value byte 2
            0x00, // value byte 3
            0x00, // value byte 4
            0x00, // value byte 5
            0x00, // value byte 6
            0x00, // value byte 7
        ];
        let len = buffer.len() as u16;
        let message =
            Response::read_from_payload(&mut buffer.as_slice(), len)
                .unwrap();
        assert_eq!(
            message,
            Response {
                error: ErrorCode::InvalidState, // 0x0107
                value: 0x6543u16.into(),
            }
        );
    }
}
